import base64

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from .constants import SALT, PASSWORD


class Cryptographer(object):

    _fernet = Fernet(base64.urlsafe_b64encode(PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=SALT,
        iterations=100000,
        backend=default_backend()
    ).derive(PASSWORD)))

    @classmethod
    def encrypted(cls, content):
        return cls._fernet.encrypt(content)

    @classmethod
    def decrypted(cls, content):
        return cls._fernet.decrypt(content)



import os

import requests

import magic
from django.conf import settings
from django.core.validators import URLValidator, ValidationError
from django.http import Http404, HttpResponse
from django.views.generic import View

from .crypt import Cryptographer


class FetchView(View):
    """
    This is a generic, insecure view that effectively undoes any security made
    available via this module.  To make it useful, you have to subclass it and
    make use of your own rules or simply rely on the auth mixins provided by
    Django:

      from django.contrib.auth.mixins import LoginRequiredMixin
      from django_encrypted_fields.views import FetchView as BaseFetchView

      class FetchView(LoginRequiredMixin, BaseFetchView):
          pass

    Using LoginRequiredMixin would effectively allow anyone with a site login
    to view *all* files, while using something like StaffRequiredMixin would
    mean that only staff members could read the file.

    Theoretically you could also write your view to be smart enough to take the
    requested path and match it against a list of permissions, allowing you to
    set out per-user permissions whilst still only using one encryption key for
    the whole site.

    """

    def get(self, request, *args, **kwargs):

        path = kwargs.get("path")

        # No path?  You're boned.  Move along.
        if not path:
            raise Http404

        if self._is_url(path):

            content = requests.get(path, stream=True).raw.read()

        else:

            # Normalise the path to strip out naughty attempts
            path = os.path.normpath(path).replace(
                settings.MEDIA_URL, settings.MEDIA_ROOT, 1)

            # Evil path request!
            if not path.startswith(settings.MEDIA_ROOT):
                raise Http404

            # The file requested doesn't exist locally.  A legit 404
            if not os.path.exists(path):
                raise Http404

            with open(path, "rb") as f:
                content = f.read()

        content = Cryptographer.decrypted(content)
        return HttpResponse(
            content, content_type=magic.Magic(mime=True).from_buffer(content))

    @staticmethod
    def _is_url(path):
        try:
            URLValidator()(path)
            return True
        except ValidationError:
            return False



from io import BytesIO

from django.conf import settings
from django.core.urlresolvers import reverse
from django.db.models.fields.files import (
    FieldFile,
    FileField,
    ImageField,
    ImageFieldFile
)

from .constants import FETCH_URL_NAME
from .crypt import Cryptographer

try:
    from urllib.parse import quote as url_encode  # Python 3
except ImportError:
    from urllib import quote as url_encode  # Python 2


class EncryptedFile(BytesIO):
    def __init__(self, content):
        self.size = content.size
        BytesIO.__init__(self, Cryptographer.encrypted(content.file.read()))


class EncryptionMixin(object):

    def save(self, name, content, save=True):
        return FieldFile.save(
            self,
            name,
            EncryptedFile(content),
            save=save
        )
    save.alters_data = True

    def _get_url(self):
        return reverse(FETCH_URL_NAME, kwargs={
            "path": super(EncryptionMixin, self).url
        })
    url = property(_get_url)


class EncryptedFieldFile(EncryptionMixin, FieldFile):
    pass


class EncryptedImageFieldFile(EncryptionMixin, ImageFieldFile):
    pass


class EncryptedFileField(FileField):
    attr_class = EncryptedFieldFile


class EncryptedImageField(ImageField):

    attr_class = EncryptedImageFieldFile

    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        """
        Since we're encrypting the file, any attempts to force recalculation of
        the dimensions will always fail, resulting in a null value for height
        and width.  To avoid that, we just set force=False all the time and
        expect that if you want to change those values, you'll do it on your
        own.
        """
        ImageField.update_dimension_fields(
            self, instance, force=False, *args, **kwargs)



import os
from django.conf import settings
from django.utils import six


def _get_setting(name):
    setting_name = "DEFF_{}".format(name)
    return os.getenv(setting_name, getattr(settings, setting_name, None))


def get_bytes(v):

    if isinstance(v, six.string_types):
        return bytes(v.encode("utf-8"))

    if isinstance(v, bytes):
        return v

    raise TypeError(
        "SALT & PASSWORD must be specified as strings that convert nicely to "
        "bytes."
    )


SALT = get_bytes(_get_setting("SALT"))
PASSWORD = get_bytes(_get_setting("PASSWORD"))
FETCH_URL_NAME = _get_setting("FETCH_URL_NAME")



import os
import re

from setuptools import setup


def get_version():
    version_file = os.path.join(
        os.path.dirname(__file__),
        'django_encrypted_filefield',
        '__init__.py'
    )
    with open(version_file) as f:
        regex = re.compile("^__version__.*?\(([\d]+), *([\d]+), *([\d]+)\).*")
        for line in f:
            m = regex.match(line)
            if m:
                return ".".join(m.groups())


def get_requirements():
    with open("requirements.txt") as f:
        requirements = []
        regex = re.compile("^([^ ]+) *.*")
        for line in f:
            if not line.startswith("#"):
                m = regex.match(line)
                if m:
                    requirements.append(line.strip())
        return requirements


# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

readme = os.path.join(os.path.dirname(__file__), "README.rst")
github = "https://github.com/danielquinn"
with open(readme) as description:
    setup(
        name="django-encrypted-filefield",
        version=get_version(),
        license="GPLv3",
        description=(
            "Encrypt uploaded files, store them wherever you like and stream "
            "them back unencrypted"
        ),
        long_description=description.read(),
        url="{}/django-encrypted-filefield".format(github),
        download_url="{}/django-encrypted-filefield".format(github),
        author="Daniel Quinn",
        author_email="code@danielquinn.org",
        maintainer="Daniel Quinn",
        maintainer_email="code@danielquinn.org",
        packages=["django_encrypted_filefield"],
        install_requires=get_requirements(),
        classifiers=[
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Topic :: Internet :: WWW/HTTP",
        ],
    )



import os

from django.core.checks import Error, register
from django.core.urlresolvers import reverse, NoReverseMatch

from .constants import FETCH_URL_NAME, PASSWORD, SALT


@register()
def constants_check(app_configs, **kwargs):

    check_messages = []

    message = "{} must be defined in your environment for " \
              "django-encrypted-filefield to work."

    if not SALT:
        check_messages.append(Error(message.format("DEFF_SALT")))
    if not PASSWORD:
        check_messages.append(Error(message.format("DEFF_PASSWORD")))

    return check_messages


@register()
def fetch_url_check(app_configs, **kwargs):

    if not FETCH_URL_NAME:
        return []  # We've got bigger problems

    try:
        reverse(FETCH_URL_NAME, kwargs={"path": "anything"})
    except NoReverseMatch:
        return [Error(
            "django-encrypted-filefield requires that you define a url for "
            "the fetching the files."
        )]

    return []
